@@ -1,5 +1,5 @@
 /*
-  Copyright (c) 2005 by Jakob Schroeter <js@camaya.net>
+  Copyright (c) 2005-2006 by Jakob Schroeter <js@camaya.net>
   This file is part of the gloox library. http://camaya.net/gloox
 
   This software is distributed under a license. The full license
@@ -20,9 +20,20 @@
 
 #include "clientbase.h"
 #include "connection.h"
+#include "messagesessionhandler.h"
 #include "parser.h"
 #include "tag.h"
 #include "stanza.h"
+#include "connectionlistener.h"
+#include "iqhandler.h"
+#include "messagehandler.h"
+#include "presencehandler.h"
+#include "rosterlistener.h"
+#include "subscriptionhandler.h"
+#include "loghandler.h"
+#include "taghandler.h"
+#include "jid.h"
+#include "base64.h"
 
 #include <iksemel.h>
 
@@ -36,18 +47,22 @@ namespace gloox
 
   ClientBase::ClientBase( const std::string& ns, const std::string& server, int port )
     : m_connection( 0 ), m_namespace( ns ), m_xmllang( "en" ), m_server( server ),
-      m_authed( false ), m_sasl( true ), m_tls( true ), m_port( port ), m_parser( 0 ),
-      m_authError( AUTH_ERROR_UNDEFINED ), m_streamError( ERROR_UNDEFINED ),
-      m_streamErrorAppCondition( 0 ), m_idCount( 0 )
+      m_authed( false ), m_sasl( true ), m_tls( true ), m_port( port ),
+      m_messageSessionHandler( 0 ), m_parser( 0 ),
+      m_authError( AuthErrorUndefined ), m_streamError( StreamErrorUndefined ),
+      m_streamErrorAppCondition( 0 ), m_idCount( 0 ), m_autoMessageSession( false ),
+      m_fdRequested( false )
   {
   }
 
   ClientBase::ClientBase( const std::string& ns, const std::string& password,
                           const std::string& server, int port )
     : m_connection( 0 ), m_namespace( ns ), m_password( password ), m_xmllang( "en" ), m_server( server ),
-      m_authed( false ), m_sasl( true ), m_tls( true ), m_port( port ), m_parser( 0 ),
-      m_authError( AUTH_ERROR_UNDEFINED ), m_streamError( ERROR_UNDEFINED ),
-      m_streamErrorAppCondition( 0 ), m_idCount( 0 )
+      m_authed( false ), m_sasl( true ), m_tls( true ), m_port( port ),
+      m_messageSessionHandler( 0 ), m_parser( 0 ),
+      m_authError( AuthErrorUndefined ), m_streamError( StreamErrorUndefined ),
+      m_streamErrorAppCondition( 0 ), m_idCount( 0 ), m_autoMessageSession( false ),
+      m_fdRequested( false )
   {
   }
 
@@ -59,8 +74,11 @@ namespace gloox
 
   ConnectionError ClientBase::recv( int timeout )
   {
+    if( !m_connection || m_connection->state() == StateDisconnected )
+      return ConnNotConnected;
+
     ConnectionError e = m_connection->recv( timeout );
-    if( e != CONN_OK )
+    if( e != ConnNoError )
       notifyOnDisconnect( e );
     return e;
   }
@@ -74,13 +92,16 @@ namespace gloox
       m_parser = new Parser( this );
 
     if( !m_connection )
-      m_connection = new Connection( m_parser, m_server, m_port );
+      m_connection = new Connection( m_parser, m_logInstance, m_server, m_port );
 
 #ifdef HAVE_TLS
     m_connection->setCACerts( m_cacerts );
+    if( !m_clientKey.empty() && !m_clientCerts.empty() )
+      m_connection->setClientCert( m_clientKey, m_clientCerts );
 #endif
+
     int ret = m_connection->connect();
-    if( ret == STATE_CONNECTED )
+    if( ret == StateConnected )
     {
       header();
       if( block )
@@ -99,22 +120,18 @@ namespace gloox
   void ClientBase::filter( NodeType type, Stanza *stanza )
   {
     if( stanza )
-      log( stanza->xml(), true );
+      logInstance().log( LogLevelDebug, LogAreaXmlIncoming, stanza->xml() );
 
     switch( type )
     {
       case NODE_STREAM_START:
       {
         const std::string version = stanza->findAttribute( "version" );
-        if( !version.empty() )
+        if( !checkStreamVersion( version ) )
         {
-          if( !checkStreamVersion( version ) )
-            disconnect( CONN_STREAM_ERROR );
-        }
-        else
-        {
-          printf( "This server is not XMPP-compliant (it does not send a 'version' attribute). Please try another one.\n" );
-          disconnect( CONN_STREAM_ERROR );
+          logInstance().log( LogLevelDebug, LogAreaClassClientbase, "This server is not XMPP-compliant"
+              " (it does not send a 'version' attribute). Please fix it or try another one.\n" );
+          disconnect( ConnStreamError );
         }
 
         m_sid = stanza->findAttribute( "id" );
@@ -126,16 +143,16 @@ namespace gloox
         {
           switch( stanza->type() )
           {
-            case STANZA_IQ:
+            case StanzaIq:
               notifyIqHandlers( stanza );
               break;
-            case STANZA_PRESENCE:
+            case StanzaPresence:
               notifyPresenceHandlers( stanza );
               break;
-            case STANZA_S10N:
+            case StanzaS10n:
               notifySubscriptionHandlers( stanza );
               break;
-            case STANZA_MESSAGE:
+            case StanzaMessage:
               notifyMessageHandlers( stanza );
               break;
             default:
@@ -146,42 +163,39 @@ namespace gloox
         break;
       case NODE_STREAM_ERROR:
         handleStreamError( stanza );
+        disconnect( ConnStreamError );
+        if( m_fdRequested )
+          notifyOnDisconnect( ConnStreamError );
         break;
       case NODE_STREAM_CLOSE:
-#ifdef DEBUG
-        printf( "stream closed\n" );
-#endif
-        disconnect( CONN_STREAM_CLOSED );
+        logInstance().log( LogLevelDebug, LogAreaClassClientbase, "stream closed" );
+        disconnect( ConnStreamClosed );
         break;
     }
   }
 
-  void ClientBase::disconnect()
-  {
-    disconnect( CONN_USER_DISCONNECTED );
-  }
-
   void ClientBase::disconnect( ConnectionError reason )
   {
     if( m_connection )
     {
-      if( reason == CONN_USER_DISCONNECTED )
-        m_streamError = ERROR_UNDEFINED;
+      if( reason == ConnUserDisconnected )
+        m_streamError = StreamErrorUndefined;
       m_connection->disconnect( reason );
     }
   }
 
   void ClientBase::header()
   {
-    std::string xml = "<?xml version='1.0' ?>";
-    xml += "<stream:stream to='" + m_jid.server()+ "' xmlns='" + m_namespace + "' ";
-    xml += "xmlns:stream='http://etherx.jabber.org/streams'  xml:lang='" + m_xmllang + "' ";
-    xml += "version='";
-    xml += XMPP_STREAM_VERSION_MAJOR;
-    xml += ".";
-    xml += XMPP_STREAM_VERSION_MINOR;
-    xml += "'>";
-    send( xml );
+    std::ostringstream oss;
+    oss << "<?xml version='1.0' ?>";
+    oss << "<stream:stream to='" + m_jid.server()+ "' xmlns='" + m_namespace + "' ";
+    oss << "xmlns:stream='http://etherx.jabber.org/streams'  xml:lang='" + m_xmllang + "' ";
+    oss << "version='";
+    oss << XMPP_STREAM_VERSION_MAJOR;
+    oss << ".";
+    oss << XMPP_STREAM_VERSION_MINOR;
+    oss << "'>";
+    send( oss.str() );
   }
 
   bool ClientBase::hasTls()
@@ -197,39 +211,47 @@ namespace gloox
   void ClientBase::startTls()
   {
     Tag *start = new Tag( "starttls" );
-    start->addAttrib( "xmlns", XMLNS_STREAM_TLS );
+    start->addAttribute( "xmlns", XMLNS_STREAM_TLS );
     send( start );
   }
+
+  void ClientBase::setClientCert( const std::string& clientKey, const std::string& clientCerts )
+  {
+    m_clientKey = clientKey;
+    m_clientCerts = clientCerts;
+  }
 #endif
 
   void ClientBase::startSASL( SaslMechanisms type )
   {
     Tag *a = new Tag( "auth" );
-    a->addAttrib( "xmlns", XMLNS_STREAM_SASL );
+    a->addAttribute( "xmlns", XMLNS_STREAM_SASL );
 
     switch( type )
     {
-      case SASL_DIGEST_MD5:
-        a->addAttrib( "mechanism", "DIGEST-MD5" );
+      case SaslDigestMd5:
+        a->addAttribute( "mechanism", "DIGEST-MD5" );
         break;
-      case SASL_PLAIN:
+      case SaslPlain:
       {
-        a->addAttrib( "mechanism", "PLAIN" );
-        int len = m_jid.username().length() + m_password.length() + 2;
-        char *tmp = (char*)iks_malloc( len + 80 );
-        char *result;
+        a->addAttribute( "mechanism", "PLAIN" );
+        size_t len = m_jid.username().length() + m_password.length() + 2;
+        char *tmp = (char*)malloc( len + 80 );
         sprintf( tmp, "%c%s%c%s", 0, m_jid.username().c_str(), 0, m_password.c_str() );
-        result = iks_base64_encode( tmp, len );
-
-        a->setCData( result );
-        iks_free( result );
-        iks_free( tmp );
+        std::string dec;
+        dec.assign( tmp, len );
+        a->setCData( Base64::encode64( dec ) );
+        free( tmp );
         break;
       }
-      case SASL_ANONYMOUS:
-        a->addAttrib( "mechanism", "ANONYMOUS" );
+      case SaslAnonymous:
+        a->addAttribute( "mechanism", "ANONYMOUS" );
         a->setCData( getID() );
         break;
+      case SaslExternal:
+        a->addAttribute( "mechanism", "EXTERNAL" );
+        a->setCData( Base64::encode64( m_jid.bare() ) );
+        break;
     }
 
     send( a );
@@ -240,11 +262,7 @@ namespace gloox
     const int CNONCE_LEN = 4;
     Tag *t;
     std::string decoded, nonce, realm, response;
-    char *b = iks_base64_decode( challenge.c_str() );
-    if( b )
-      decoded = b;
-    else
-      return;
+    decoded = Base64::decode64( challenge.c_str() );
 
     if( decoded.substr( 0, 7 ) == "rspauth" )
     {
@@ -255,30 +273,28 @@ namespace gloox
       char cnonce[CNONCE_LEN*8 + 1];
       unsigned char a1_h[16];
       char a1[33], a2[33], response_value[33];
-      char *response_coded;
       iksmd5 *md5;
       int i;
 
-      unsigned int r_pos = decoded.find( "realm=" );
+      size_t r_pos = decoded.find( "realm=" );
       if( r_pos != std::string::npos )
       {
-        unsigned int r_end = decoded.find( "\"", r_pos + 7 );
+        size_t r_end = decoded.find( "\"", r_pos + 7 );
         realm = decoded.substr( r_pos + 7, r_end - (r_pos + 7 ) );
       }
       else
         realm = m_jid.server();
 
-      unsigned int n_pos = decoded.find( "nonce=" );
+      size_t n_pos = decoded.find( "nonce=" );
       if( n_pos != std::string::npos )
       {
-        unsigned int n_end = decoded.find( "\"", n_pos + 7 );
+        size_t n_end = decoded.find( "\"", n_pos + 7 );
         while( decoded.substr( n_end-1, 1 ) == "\\" )
           n_end = decoded.find( "\"", n_end + 1 );
         nonce = decoded.substr( n_pos + 7, n_end - ( n_pos + 7 ) );
       }
       else
       {
-        iks_free( b );
         return;
       }
 
@@ -324,34 +340,30 @@ namespace gloox
       response += "\",nc=00000001,qop=auth,digest-uri=\"xmpp/" + m_jid.server() + "\",response=";
       response += response_value;
       response += ",charset=utf-8";
-      response_coded = iks_base64_encode( response.c_str(), response.length() );
 
-      t = new Tag( "response", response_coded );
-
-      iks_free( response_coded );
+      t = new Tag( "response", Base64::encode64( response ) );
     }
-    t->addAttrib( "xmlns", XMLNS_STREAM_SASL );
+    t->addAttribute( "xmlns", XMLNS_STREAM_SASL );
     send( t );
-    iks_free( b );
 
   }
 
   void ClientBase::processSASLError( Stanza *stanza )
   {
     if( stanza->hasChild( "aborted" ) )
-      m_authError = SASL_ABORTED;
+      m_authError = SaslAborted;
     else if( stanza->hasChild( "incorrect-encoding" ) )
-      m_authError = SASL_INCORRECT_ENCODING;
+      m_authError = SaslIncorrectEncoding;
     else if( stanza->hasChild( "invalid-authzid" ) )
-      m_authError = SASL_INVALID_AUTHZID;
+      m_authError = SaslInvalidAuthzid;
     else if( stanza->hasChild( "invalid-mechanism" ) )
-      m_authError = SASL_INVALID_MECHANISM;
+      m_authError = SaslInvalidMechanism;
     else if( stanza->hasChild( "mechanism-too-weak" ) )
-      m_authError = SASL_MECHANISM_TOO_WEAK;
+      m_authError = SaslMechanismTooWeak;
     else if( stanza->hasChild( "not-authorized" ) )
-      m_authError = SASL_NOT_AUTHORIZED;
+      m_authError = SaslNotAuthorized;
     else if( stanza->hasChild( "temporary-auth-failure" ) )
-      m_authError = SASL_TEMPORARY_AUTH_FAILURE;
+      m_authError = SaslTemporaryAuthFailure;
   }
 
   void ClientBase::send( Tag *tag )
@@ -361,18 +373,18 @@ namespace gloox
 
     send( tag->xml() );
 
-    if( tag->type() == STANZA_UNDEFINED )
-      delete( tag );
+    if( tag->type() == StanzaUndefined )
+      delete tag;
     else
     {
       Stanza *s = dynamic_cast<Stanza*>( tag );
-      delete( s );
+      delete s;
     }
   }
 
   void ClientBase::send( const std::string& xml )
   {
-    log( xml, false );
+    logInstance().log( LogLevelDebug, LogAreaXmlOutgoing, xml );
 
     if( m_connection )
       m_connection->send( xml );
@@ -382,7 +394,7 @@ namespace gloox
     if( m_connection )
       return m_connection->state();
     else
-      return STATE_DISCONNECTED;
+      return StateDisconnected;
   }
 
   const std::string ClientBase::getID()
@@ -394,11 +406,14 @@ namespace gloox
 
   bool ClientBase::checkStreamVersion( const std::string& version )
   {
+    if( version.empty() )
+      return false;
+
     int major = 0;
     int minor = 0;
-    int myMajor = atoi( XMPP_STREAM_VERSION_MAJOR );
+    int myMajor = XMPP_STREAM_VERSION_MAJOR;
 
-    unsigned int dot = version.find( "." );
+    size_t dot = version.find( "." );
     if( !version.empty() && dot && dot != std::string::npos )
     {
       major = atoi( version.substr( 0, dot ).c_str() );
@@ -411,6 +426,11 @@ namespace gloox
       return true;
   }
 
+  LogSink& ClientBase::logInstance()
+  {
+    return m_logInstance;
+  }
+
   void ClientBase::handleStreamError( Stanza *stanza )
   {
     Tag::TagList& c = stanza->children();
@@ -418,58 +438,58 @@ namespace gloox
     for( ; it != c.end(); ++it )
     {
       if( (*it)->name() == "bad-format" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_BAD_FORMAT;
+        m_streamError = StreamErrorBadFormat;
       else if( (*it)->name() == "bad-namespace-prefix" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_BAD_NAMESPACE_PREFIX;
+        m_streamError = StreamErrorBadNamespacePrefix;
       else if( (*it)->name() == "conflict" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_CONFLICT;
+        m_streamError = StreamErrorConflict;
       else if( (*it)->name() == "connection-timeout" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_CONNECTION_TIMEOUT;
+        m_streamError = StreamErrorConnectionTimeout;
       else if( (*it)->name() == "host-gone" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_HOST_GONE;
+        m_streamError = StreamErrorHostGone;
       else if( (*it)->name() == "host-unknown" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_HOST_UNKNOWN;
+        m_streamError = StreamErrorHostUnknown;
       else if( (*it)->name() == "improper-addressing" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_IMPROPER_ADDRESSING;
+        m_streamError = StreamErrorImproperAddressing;
       else if( (*it)->name() == "internal-server-error" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_INTERNAL_SERVER_ERROR;
+        m_streamError = StreamErrorInternalServerError;
       else if( (*it)->name() == "invalid-from" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_INVALID_FROM;
+        m_streamError = StreamErrorInvalidFrom;
       else if( (*it)->name() == "invalid-id" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_INVALID_ID;
+        m_streamError = StreamErrorInvalidId;
       else if( (*it)->name() == "invalid-namespace" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_INVALID_NAMESPACE;
+        m_streamError = StreamErrorInvalidNamespace;
       else if( (*it)->name() == "invalid-xml" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_INVALID_XML;
+        m_streamError = StreamErrorInvalidXml;
       else if( (*it)->name() == "not-authorized" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_NOT_AUTHORIZED;
+        m_streamError = StreamErrorNotAuthorized;
       else if( (*it)->name() == "policy-violation" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_POLICY_VIOLATION;
+        m_streamError = StreamErrorPolicyViolation;
       else if( (*it)->name() == "remote-connection-failed" &&
                  (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_REMOTE_CONNECTION_FAILED;
+        m_streamError = StreamErrorRemoteConnectionFailed;
       else if( (*it)->name() == "resource-constraint" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_RESOURCE_CONSTRAINT;
+        m_streamError = StreamErrorResourceConstraint;
       else if( (*it)->name() == "restricted-xml" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_RESTRICTED_XML;
+        m_streamError = StreamErrorRestrictedXml;
       else if( (*it)->name() == "see-other-host" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
       {
-        m_streamError = ERROR_SEE_OTHER_HOST;
+        m_streamError = StreamErrorSeeOtherHost;
         m_streamErrorCData = stanza->findChild( "see-other-host" )->cdata();
       }
       else if( (*it)->name() == "system-shutdown" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_SYSTEM_SHUTDOWN;
+        m_streamError = StreamErrorSystemShutdown;
       else if( (*it)->name() == "undefined-condition" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_UNDEFINED_CONDITION;
+        m_streamError = StreamErrorUndefinedCondition;
       else if( (*it)->name() == "unsupported-encoding" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_UNSUPPORTED_ENCODING;
+        m_streamError = StreamErrorUnsupportedEncoding;
       else if( (*it)->name() == "unsupported-stanza-type" &&
                  (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_UNSUPPORTED_STANZA_TYPE;
+        m_streamError = StreamErrorUnsupportedStanzaType;
       else if( (*it)->name() == "unsupported-version" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_UNSUPPORTED_VERSION;
+        m_streamError = StreamErrorUnsupportedVersion;
       else if( (*it)->name() == "xml-not-well-formed" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
-        m_streamError = ERROR_XML_NOT_WELL_FORMED;
+        m_streamError = StreamErrorXmlNotWellFormed;
       else if( (*it)->name() == "text" && (*it)->hasAttribute( "xmlns", XMLNS_XMPP_STREAM ) )
       {
         const std::string lang = (*it)->findAttribute( "xml:lang" );
@@ -481,8 +501,6 @@ namespace gloox
       else
         m_streamErrorAppCondition = (*it);
     }
-
-    disconnect( CONN_STREAM_ERROR );
   }
 
   const std::string ClientBase::streamErrorText( const std::string& lang ) const
@@ -494,28 +512,27 @@ namespace gloox
       return "";
   }
 
-  void ClientBase::log( const std::string& xml, bool incoming )
+  void ClientBase::setAutoMessageSession( bool autoMS, MessageSessionHandler *msh )
   {
-#ifdef DEBUG
-    if ( m_connection->isSecure() )
-      printf( "Sec" );
-
-    if( incoming )
-      printf( "RECV " );
+    if( autoMS && msh )
+    {
+      m_messageSessionHandler = msh;
+      m_autoMessageSession = true;
+    }
     else
-      printf( "SEND " );
-
-    printf( "[%s]", xml.c_str() );
-    if( xml.substr( xml.length()-2, 1 ) != "\n" )
-      printf( "\n" );
-#endif
-    notifyLogHandlers( xml, incoming );
+    {
+      m_autoMessageSession = false;
+      m_messageSessionHandler = 0;
+    }
   }
 
-  int ClientBase::fileDescriptor() const
+  int ClientBase::fileDescriptor()
   {
     if( m_connection )
+    {
+      m_fdRequested = true;
       return m_connection->fileDescriptor();
+    }
     else
       return -1;
   }
@@ -532,16 +549,6 @@ namespace gloox
       m_presenceHandlers.remove( ph );
   }
 
-  void ClientBase::registerLogHandler( LogHandler *lh )
-  {
-    m_logHandlers.push_back( lh );
-  }
-
-  void ClientBase::removeLogHandler( LogHandler *lh )
-  {
-    m_logHandlers.remove( lh );
-  }
-
   void ClientBase::registerIqHandler( IqHandler *ih, const std::string& xmlns )
   {
     if( ih && !xmlns.empty() )
@@ -565,12 +572,25 @@ namespace gloox
       m_iqNSHandlers.erase( xmlns );
   }
 
+  void ClientBase::registerMessageHandler( const std::string& jid, MessageHandler *mh )
+  {
+    if( mh && !jid.empty() )
+      m_messageJidHandlers[jid] = mh;
+  }
+
   void ClientBase::registerMessageHandler( MessageHandler *mh )
   {
     if( mh )
       m_messageHandlers.push_back( mh );
   }
 
+  void ClientBase::removeMessageHandler( const std::string& jid )
+  {
+    MessageHandlerMap::iterator it = m_messageJidHandlers.find( jid );
+    if( it != m_messageJidHandlers.end() )
+      m_messageJidHandlers.erase( it );
+  }
+
   void ClientBase::removeMessageHandler( MessageHandler *mh )
   {
     if( mh )
@@ -642,6 +662,8 @@ namespace gloox
     {
       (*it)->onDisconnect( e );
     }
+
+    cleanup();
   }
 
   bool ClientBase::notifyOnTLSConnect( const CertInfo& info )
@@ -649,13 +671,14 @@ namespace gloox
     ConnectionListenerList::const_iterator it = m_connectionListeners.begin();
     for( ; it != m_connectionListeners.end(); ++it )
     {
-      return (*it)->onTLSConnect( info );
+      if( !(*it)->onTLSConnect( info ) )
+        return false;
     }
 
-    return false;
+    return true;
   }
 
-  void ClientBase::notifyOnResourceBindError( ConnectionListener::ResourceBindError error )
+  void ClientBase::notifyOnResourceBindError( ResourceBindError error )
   {
     ConnectionListenerList::const_iterator it = m_connectionListeners.begin();
     for( ; it != m_connectionListeners.end(); ++it )
@@ -664,7 +687,7 @@ namespace gloox
     }
   }
 
-  void ClientBase::notifyOnSessionCreateError( ConnectionListener::SessionCreateError error )
+  void ClientBase::notifyOnSessionCreateError( SessionCreateError error )
   {
     ConnectionListenerList::const_iterator it = m_connectionListeners.begin();
     for( ; it != m_connectionListeners.end(); ++it )
@@ -713,19 +736,34 @@ namespace gloox
       m_iqIDHandlers.erase( it_id );
     }
 
-    if( !res && ( stanza->type() == STANZA_IQ ) &&
-         ( ( stanza->subtype() == STANZA_IQ_GET ) || ( stanza->subtype() == STANZA_IQ_SET ) ) )
+    if( !res && ( stanza->type() == StanzaIq ) &&
+         ( ( stanza->subtype() == StanzaIqGet ) || ( stanza->subtype() == StanzaIqSet ) ) )
     {
       Tag *iq = new Tag( "iq" );
-      iq->addAttrib( "type", "result" );
-      iq->addAttrib( "id", stanza->id() );
-      iq->addAttrib( "to", stanza->from().full() );
+      iq->addAttribute( "type", "result" );
+      iq->addAttribute( "id", stanza->id() );
+      iq->addAttribute( "to", stanza->from().full() );
       send( iq );
     }
   }
 
   void ClientBase::notifyMessageHandlers( Stanza *stanza )
   {
+    MessageHandlerMap::const_iterator it1 = m_messageJidHandlers.find( stanza->from().full() );
+    if( it1 != m_messageJidHandlers.end() )
+    {
+      (*it1).second->handleMessage( stanza );
+      return;
+    }
+
+    if( m_autoMessageSession && m_messageSessionHandler )
+    {
+      MessageSession *session = new MessageSession( this, stanza->from() );
+      m_messageSessionHandler->handleMessageSession( session );
+      notifyMessageHandlers( stanza );
+      return;
+    }
+
     MessageHandlerList::const_iterator it = m_messageHandlers.begin();
     for( ; it != m_messageHandlers.end(); ++it )
     {
@@ -733,15 +771,6 @@ namespace gloox
     }
   }
 
-  void ClientBase::notifyLogHandlers( const std::string& xml, bool incoming )
-  {
-    LogHandlerList::const_iterator it = m_logHandlers.begin();
-    for( ; it != m_logHandlers.end(); ++it )
-    {
-      (*it)->handleLog( xml, incoming );
-    }
-  }
-
   void ClientBase::notifyTagHandlers( Stanza *stanza )
   {
     TagHandlerList::const_iterator it = m_tagHandlers.begin();
@@ -752,4 +781,9 @@ namespace gloox
     }
   }
 
+  void ClientBase::cleanup()
+  {
+
+  }
+
 }
