@@ -1,5 +1,5 @@
 /*
-  Copyright (c) 2005 by Jakob Schroeter <js@camaya.net>
+  Copyright (c) 2005-2006 by Jakob Schroeter <js@camaya.net>
   This file is part of the gloox library. http://camaya.net/gloox
 
   This software is distributed under a license. The full license
@@ -15,18 +15,10 @@
 #ifndef CLIENTBASE_H__
 #define CLIENTBASE_H__
 
+#include "macros.h"
 #include "gloox.h"
-
-#include "connectionlistener.h"
-#include "iqhandler.h"
-#include "messagehandler.h"
-#include "presencehandler.h"
-#include "rosterlistener.h"
-#include "subscriptionhandler.h"
-#include "loghandler.h"
-#include "taghandler.h"
 #include "jid.h"
-
+#include "logsink.h"
 
 namespace gloox
 {
@@ -39,21 +31,30 @@ namespace gloox
   class Tag;
   class Stanza;
   class Parser;
+  class MessageSessionHandler;
+  class ConnectionListener;
+  class IqHandler;
+  class MessageHandler;
+  class PresenceHandler;
+//   class RosterListener;
+  class SubscriptionHandler;
+//   class LogHandler;
+  class TagHandler;
 
   /**
    * @brief This is the common base class for a jabber Client and a jabber Component.
    *
-   * It manages connection establishing, authentication, filter registrationa and invocation.
+   * It manages connection establishing, authentication, filter registration and invocation.
    * @author Jakob Schroeter <js@camaya.net>
    * @since 0.3
    */
-  class GLOOX_EXPORT ClientBase
+  class GLOOX_API ClientBase
   {
 
     friend class Parser;
+    friend class RosterManager;
 
     public:
-
       /**
        * Constructs a new ClientBase.
        * You should not need to use this class directly. Use Client or Component instead.
@@ -96,16 +97,12 @@ namespace gloox
       /**
        * Use this periodically to receive data from the socket and to feed the parser. You need to use
        * this only if you chose to connect in non-blocking mode.
-       * @param timeout The timeout to use for select. Default of -1 means blocking until data was available.
+       * @param timeout The timeout in seconds to use for select. Default of -1 means blocking
+       * until data was available.
        * @return The state of the connection.
        */
       ConnectionError recv( int timeout = -1 );
 
-      /**
-       * Disconnects from the server.
-       */
-      void disconnect();
-
       /**
        * Reimplement this function to provide a username for connection purposes.
        * @return The username.
@@ -193,6 +190,8 @@ namespace gloox
 
       /**
        * Sends a given Tag over an established connection.
+       * The ClientBase object becomes the owner of this Tag and will delete it after sending it.
+       * You should not rely on the existance of the Tag after it's been sent.
        * @param tag The Tag to send.
        */
       virtual void send( Tag *tag );
@@ -229,7 +228,7 @@ namespace gloox
        * can use select() on it and use Connection::recv( -1 ) to fetch the data.
        * @return The file descriptor of the active connection, or -1 if no connection is established.
        */
-      int fileDescriptor() const;
+      int fileDescriptor();
 
       /**
        * Registers @c cl as object that receives connection notifications.
@@ -261,6 +260,15 @@ namespace gloox
        */
       void registerMessageHandler( MessageHandler *mh );
 
+      /**
+       * Registers @c mh as object that receives Message stanza notifications for the given JID.
+       * Only one message handler per (full) JID is possible.
+       * You should not need to use this functionality directly. Instead use a MessageSession.
+       * @param jid Messages from this full JID will be sent to the given MessageHandler.
+       * @param mh The object to receive Message stanza notifications.
+       */
+      void registerMessageHandler( const std::string& jid,  MessageHandler *mh );
+
       /**
        * Registers @c ph as object that receives Presence stanza notifications.
        * @param ph The object to receive Presence stanza notifications.
@@ -273,13 +281,6 @@ namespace gloox
        */
       void registerSubscriptionHandler( SubscriptionHandler *sh );
 
-      /**
-       * Registers @c lh as object that receives all XML sent back and forth on the connection.
-       * Suitable for logging to a file, etc.
-       * @param lh The object to receive exchanged data.
-       */
-      void registerLogHandler( LogHandler *lh );
-
       /**
        * Registers @c th as object that receives incoming packts with a given root tag
        * qualified by the given namespace.
@@ -308,6 +309,12 @@ namespace gloox
        */
       void removeMessageHandler( MessageHandler *mh );
 
+      /**
+       * Removes the message handler for the given JID.
+       * @param jid The JID for which listening shall end.
+       */
+      void removeMessageHandler( const std::string& jid );
+
       /**
        * Removes the given object from the list of presence handlers.
        * @param ph The object to remove from the list.
@@ -330,18 +337,43 @@ namespace gloox
                                               const std::string& xmlns );
 
       /**
-       * Removes the given object from the list of log handlers.
-       * @param lh The object to remove from the list.
-       */
-      void removeLogHandler( LogHandler *lh );
-
-      /**
-       * Use this function to set a number of trusted root CA certificates. which shall be
+       * Use this function to set a number of trusted root CA certificates which shall be
        * used to verify a servers certificate.
        * @param cacerts A list of absolute paths to CA root certificate files in PEM format.
        */
       void setCACerts( const StringList& cacerts ) { m_cacerts = cacerts; };
 
+      /**
+       * Use this function to set the user's certificate and private key. The certificate will
+       * be presented to the server upon request and can be used for SASL EXTERNAL authentication.
+       * The user's certificate file should be a bundle of more than one certificate in PEM format.
+       * The first one in the file should be the user's certificate, each cert following that one
+       * should have signed the previous one.
+       * @note These certificates are not necessarily the same as those used to verify the server's
+       * certificate.
+       * @param clientKey The absolute path to the user's private key in PEM format.
+       * @param clientCerts A path to a certificate bundle in PEM format.
+       */
+      void setClientCert( const std::string& clientKey, const std::string& clientCerts );
+
+      /**
+       * Use this function to turn the Auto-MessageSession feature on or off.
+       * If this is enabled, a MessageSession will be created for every incoming
+       * message stanza if there is no MessageHandler registered for the originating JID.
+       * If you disable automatic MessageSession creation, the MessageSessionHandler will
+       * be cleared. You have to set it anew the next time you want to enable it. You cannot
+       * enable this feature without a valid MessageSessionHandler.
+       * @param autoMS Whether to enable automatic MessageSession creation.
+       * @param msh The MessageSessionHandler that will receive the newly created MessageSession.
+       */
+      void setAutoMessageSession( bool autoMS, MessageSessionHandler *msh );
+
+      /**
+       * Returns the LogSink instance for this ClientBase and all related objects.
+       * @return The LogSink instance used in the current ClientBase.
+       */
+      LogSink& logInstance();
+
       /**
        * Use this function to retrieve the type of the stream error after it occurs and you received a
        * ConnectionError of type CONN_STREAM_ERROR from the ConnectionListener.
@@ -384,18 +416,18 @@ namespace gloox
     protected:
       enum SaslMechanisms
       {
-        SASL_DIGEST_MD5,          /**< SASL Digest-MD5 according to RFC 2831. */
-        SASL_PLAIN,               /**< SASL PLAIN according to RFC 2595 Section 6. */
-        SASL_ANONYMOUS            /**< SASL ANONYMOUS according to draft-ietf-sasl-anon-05.txt/
+        SaslDigestMd5,          /**< SASL Digest-MD5 according to RFC 2831. */
+        SaslPlain,               /**< SASL PLAIN according to RFC 2595 Section 6. */
+        SaslAnonymous,           /**< SASL ANONYMOUS according to draft-ietf-sasl-anon-05.txt/
                                    * RFC 2245 Section 6. */
+        SaslExternal             /**< SASL EXTERNAL according to RFC 2222 Section 7.4. */
       };
 
-      void notifyOnResourceBindError( ConnectionListener::ResourceBindError error );
-      void notifyOnSessionCreateError( ConnectionListener::SessionCreateError error );
+      void notifyOnResourceBindError( ResourceBindError error );
+      void notifyOnSessionCreateError( SessionCreateError error );
       bool notifyOnTLSConnect( const CertInfo& info );
-      void log( const std::string& xml, bool incoming );
       void notifyOnConnect();
-      void disconnect( ConnectionError reason );
+      virtual void disconnect( ConnectionError reason );
       void header();
       void setAuthed( bool authed ) { m_authed = authed; };
       void setAuthFailure( AuthenticationError e ) { m_authError = e; };
@@ -410,6 +442,8 @@ namespace gloox
       JID m_jid;
       Connection *m_connection;
 
+      std::string m_clientCerts;
+      std::string m_clientKey;
       std::string m_namespace;
       std::string m_password;
       std::string m_xmllang;
@@ -431,6 +465,8 @@ namespace gloox
 
       virtual void handleStartNode() = 0;
       virtual bool handleNormalNode( Stanza *stanza ) = 0;
+      virtual void rosterFilled() = 0;
+      virtual void cleanup();
       void handleStreamError( Stanza *stanza );
 
       void notifyIqHandlers( Stanza *stanza );
@@ -438,7 +474,6 @@ namespace gloox
       void notifyPresenceHandlers( Stanza *stanza );
       void notifySubscriptionHandlers( Stanza *stanza );
       void notifyTagHandlers( Stanza *stanza );
-      void notifyLogHandlers( const std::string& xml, bool incoming );
       void notifyOnDisconnect( ConnectionError e );
       void filter( NodeType type, Stanza *stanza );
       void logEvent( const char *data, size_t size, int is_incoming );
@@ -457,33 +492,37 @@ namespace gloox
         std::string tag;
       };
 
-      typedef std::list<ConnectionListener*>            ConnectionListenerList;
-      typedef std::map<const std::string, IqHandler*>   IqHandlerMap;
-      typedef std::map<const std::string, TrackStruct>  IqTrackMap;
-      typedef std::list<MessageHandler*>                MessageHandlerList;
-      typedef std::list<PresenceHandler*>               PresenceHandlerList;
-      typedef std::list<SubscriptionHandler*>           SubscriptionHandlerList;
-      typedef std::list<LogHandler*>                    LogHandlerList;
-      typedef std::list<TagHandlerStruct>               TagHandlerList;
+      typedef std::list<ConnectionListener*>                ConnectionListenerList;
+      typedef std::map<const std::string, IqHandler*>       IqHandlerMap;
+      typedef std::map<const std::string, TrackStruct>      IqTrackMap;
+      typedef std::map<const std::string, MessageHandler*>  MessageHandlerMap;
+      typedef std::list<MessageHandler*>                    MessageHandlerList;
+      typedef std::list<PresenceHandler*>                   PresenceHandlerList;
+      typedef std::list<SubscriptionHandler*>               SubscriptionHandlerList;
+      typedef std::list<TagHandlerStruct>                   TagHandlerList;
 
       ConnectionListenerList  m_connectionListeners;
       IqHandlerMap            m_iqNSHandlers;
       IqTrackMap              m_iqIDHandlers;
+      MessageHandlerMap       m_messageJidHandlers;
       MessageHandlerList      m_messageHandlers;
       PresenceHandlerList     m_presenceHandlers;
       SubscriptionHandlerList m_subscriptionHandlers;
-      LogHandlerList          m_logHandlers;
       TagHandlerList          m_tagHandlers;
       StringList              m_cacerts;
+      MessageSessionHandler  *m_messageSessionHandler;
 
       Parser *m_parser;
+      LogSink m_logInstance;
 
       AuthenticationError m_authError;
       StreamError m_streamError;
-	  StringMap m_streamErrorText;
+      StringMap m_streamErrorText;
       std::string m_streamErrorCData;
       Tag *m_streamErrorAppCondition;
       int m_idCount;
+      bool m_autoMessageSession;
+      bool m_fdRequested;
 
   };
 
