@@ -1,5 +1,5 @@
 /*
-  Copyright (c) 2005 by Jakob Schroeter <js@camaya.net>
+  Copyright (c) 2005-2006 by Jakob Schroeter <js@camaya.net>
   This file is part of the gloox library. http://camaya.net/gloox
 
   This software is distributed under a license. The full license
@@ -20,13 +20,20 @@
 #endif
 
 #ifdef WIN32
-#define GLOOX_EXPORT __declspec( dllexport )
-#else
-#define GLOOX_EXPORT
+#  ifdef GLOOX_EXPORTS
+#    define GLOOX_API __declspec( dllexport )
+#  else
+#    ifdef GLOOX_IMPORTS
+#      define GLOOX_API __declspec( dllimport )
+#    endif
+#  endif
+#endif
+#ifndef GLOOX_API
+#  define GLOOX_API
 #endif
 
 
-#if __GNUC__ - 0 > 3 || ( __GNUC__ - 0 == 3 && __GNUC_MINOR__ - 0 >= 2 )
+#if defined( __GNUC__ ) && ( __GNUC__ - 0 > 3 || ( __GNUC__ - 0 == 3 && __GNUC_MINOR__ - 0 >= 2 ) )
 #define GLOOX_DEPRECATED __attribute__ ( (deprecated) )
 #elif defined( _MSC_VER ) && ( _MSC_VER >= 1300 )
 #define GLOOX_DEPRECATED __declspec( deprecated )
